@@ -1,19 +1,24 @@
 from __future__ import annotations
 
 import copy
+import logging
 import warnings
-from functools import partial
-from typing import TYPE_CHECKING, Any, Callable, TypedDict
+from dataclasses import dataclass
+from functools import cached_property, partial
+from typing import TYPE_CHECKING, Any, Callable, Literal, TypedDict
 
 import numpy as np
-import numpy_groupies as npg
-from numpy.typing import DTypeLike
+from numpy.typing import ArrayLike, DTypeLike
 
 from . import aggregate_flox, aggregate_npg, xrutils
 from . import xrdtypes as dtypes
 
 if TYPE_CHECKING:
     FuncTuple = tuple[Callable | str, ...]
+    OptionalFuncTuple = tuple[Callable | str | None, ...]
+
+
+logger = logging.getLogger("flox")
 
 
 def _is_arg_reduction(func: str | Aggregation) -> bool:
@@ -30,11 +35,22 @@ class AggDtypeInit(TypedDict):
 
 
 class AggDtype(TypedDict):
+    user: DTypeLike | None
     final: np.dtype
     numpy: tuple[np.dtype | type[np.intp], ...]
     intermediate: tuple[np.dtype | type[np.intp], ...]
 
 
+def get_npg_aggregation(func, *, engine):
+    try:
+        method_ = getattr(aggregate_npg, func)
+        method = partial(method_, engine=engine)
+    except AttributeError:
+        aggregate = aggregate_npg._get_aggregate(engine).aggregate
+        method = partial(aggregate, func=func)
+    return method
+
+
 def generic_aggregate(
     group_idx,
     array,
@@ -51,17 +67,33 @@ def generic_aggregate(
         try:
             method = getattr(aggregate_flox, func)
         except AttributeError:
-            method = partial(npg.aggregate_numpy.aggregate, func=func)
-    elif engine in ["numpy", "numba"]:
+            logger.debug(f"Couldn't find {func} for engine='flox'. Falling back to numpy")
+            method = get_npg_aggregation(func, engine="numpy")
+
+    elif engine == "numbagg":
+        from . import aggregate_numbagg
+
         try:
-            method_ = getattr(aggregate_npg, func)
-            method = partial(method_, engine=engine)
+            if "var" in func or "std" in func:
+                ddof = kwargs.get("ddof", 0)
+                if aggregate_numbagg.NUMBAGG_SUPPORTS_DDOF or (ddof != 0):
+                    method = getattr(aggregate_numbagg, func)
+                else:
+                    logger.debug(f"numbagg too old for ddof={ddof}. Falling back to numpy")
+                    method = get_npg_aggregation(func, engine="numpy")
+            else:
+                method = getattr(aggregate_numbagg, func)
+
         except AttributeError:
-            aggregate = aggregate_npg._get_aggregate(engine).aggregate
-            method = partial(aggregate, func=func)
+            logger.debug(f"Couldn't find {func} for engine='numbagg'. Falling back to numpy")
+            method = get_npg_aggregation(func, engine="numpy")
+
+    elif engine in ["numpy", "numba"]:
+        method = get_npg_aggregation(func, engine=engine)
+
     else:
         raise ValueError(
-            f"Expected engine to be one of ['flox', 'numpy', 'numba']. Received {engine} instead."
+            f"Expected engine to be one of ['flox', 'numpy', 'numba', 'numbagg']. Received {engine} instead."
         )
 
     group_idx = np.asarray(group_idx, like=array)
@@ -109,12 +141,31 @@ def _get_fill_value(dtype, fill_value):
     return fill_value
 
 
-def _atleast_1d(inp):
+def _atleast_1d(inp, min_length: int = 1):
     if xrutils.is_scalar(inp):
-        inp = (inp,)
+        inp = (inp,) * min_length
+    assert len(inp) >= min_length
     return inp
 
 
+def returns_empty_tuple(*args, **kwargs):
+    return ()
+
+
+@dataclass
+class Dim:
+    values: ArrayLike
+    name: str | None
+
+    @cached_property
+    def is_scalar(self) -> bool:
+        return xrutils.is_scalar(self.values)
+
+    @cached_property
+    def size(self) -> int:
+        return 0 if self.is_scalar else len(self.values)  # type: ignore[arg-type]
+
+
 class Aggregation:
     def __init__(
         self,
@@ -124,13 +175,13 @@ class Aggregation:
         chunk: str | FuncTuple | None,
         combine: str | FuncTuple | None,
         preprocess: Callable | None = None,
-        aggregate: Callable | None = None,
         finalize: Callable | None = None,
         fill_value=None,
         final_fill_value=dtypes.NA,
         dtypes=None,
         final_dtype: DTypeLike | None = None,
-        reduction_type="reduce",
+        reduction_type: Literal["reduce", "argreduce"] = "reduce",
+        new_dims_func: Callable | None = None,
     ):
         """
         Blueprint for computing grouped aggregations.
@@ -173,6 +224,10 @@ class Aggregation:
             per reduction in ``chunk`` as a tuple.
         final_dtype : DType, optional
             DType for output. By default, uses dtype of array being reduced.
+        new_dims_func: Callable
+            Function that receives finalize_kwargs and returns a tupleof sizes of any new dimensions
+            added by the reduction. For e.g. quantile for q=(0.5, 0.85) adds a new dimension of size 2,
+            so returns (2,)
         """
         self.name = name
         # preprocess before blockwise
@@ -181,13 +236,11 @@ class Aggregation:
         self.reduction_type = reduction_type
         self.numpy: FuncTuple = (numpy,) if numpy else (self.name,)
         # initialize blockwise reduction
-        self.chunk: FuncTuple = _atleast_1d(chunk)
+        self.chunk: OptionalFuncTuple = _atleast_1d(chunk)
         # how to aggregate results after first round of reduction
-        self.combine: FuncTuple = _atleast_1d(combine)
+        self.combine: OptionalFuncTuple = _atleast_1d(combine)
         # simpler reductions used with the "simple combine" algorithm
-        self.simple_combine = None
-        # final aggregation
-        self.aggregate: Callable | str = aggregate if aggregate else self.combine[0]
+        self.simple_combine: OptionalFuncTuple = ()
         # finalize results (see mean)
         self.finalize: Callable | None = finalize
 
@@ -203,11 +256,22 @@ class Aggregation:
             "final": final_dtype,
             "intermediate": self._normalize_dtype_fill_value(dtypes, "dtype"),
         }
-        self.dtype: AggDtype = None  # type: ignore
+        self.dtype: AggDtype = None  # type: ignore[assignment]
 
         # The following are set by _initialize_aggregation
         self.finalize_kwargs: dict[Any, Any] = {}
-        self.min_count: int | None = None
+        self.min_count: int = 0
+        self.new_dims_func: Callable = (
+            returns_empty_tuple if new_dims_func is None else new_dims_func
+        )
+
+    @cached_property
+    def new_dims(self) -> tuple[Dim]:
+        return self.new_dims_func(**self.finalize_kwargs)
+
+    @cached_property
+    def num_new_vector_dims(self) -> int:
+        return len(tuple(dim for dim in self.new_dims if not dim.is_scalar))
 
     def _normalize_dtype_fill_value(self, value, name):
         value = _atleast_1d(value)
@@ -226,7 +290,6 @@ class Aggregation:
             self.numpy,
             self.chunk,
             self.combine,
-            self.aggregate,
             self.finalize,
             self.fill_value,
             self.dtype,
@@ -238,7 +301,6 @@ class Aggregation:
                 f"{self.name!r}, fill: {self.fill_value.values()!r}, dtype: {self.dtype}",
                 f"chunk: {self.chunk!r}",
                 f"combine: {self.combine!r}",
-                f"aggregate: {self.aggregate!r}",
                 f"finalize: {self.finalize!r}",
                 f"min_count: {self.min_count!r}",
             )
@@ -261,13 +323,7 @@ count = Aggregation(
 sum_ = Aggregation("sum", chunk="sum", combine="sum", fill_value=0)
 nansum = Aggregation("nansum", chunk="nansum", combine="sum", fill_value=0)
 prod = Aggregation("prod", chunk="prod", combine="prod", fill_value=1, final_fill_value=1)
-nanprod = Aggregation(
-    "nanprod",
-    chunk="nanprod",
-    combine="prod",
-    fill_value=1,
-    final_fill_value=dtypes.NA,
-)
+nanprod = Aggregation("nanprod", chunk="nanprod", combine="prod", fill_value=1)
 
 
 def _mean_finalize(sum_, count):
@@ -465,10 +521,38 @@ any_ = Aggregation(
     final_dtype=bool,
 )
 
-# numpy_groupies does not support median
-# And the dask version is really hard!
-# median = Aggregation("median", chunk=None, combine=None, fill_value=None)
-# nanmedian = Aggregation("nanmedian", chunk=None, combine=None, fill_value=None)
+# Support statistical quantities only blockwise
+# The parallel versions will be approximate and are hard to implement!
+median = Aggregation(
+    name="median", fill_value=dtypes.NA, chunk=None, combine=None, final_dtype=np.float64
+)
+nanmedian = Aggregation(
+    name="nanmedian", fill_value=dtypes.NA, chunk=None, combine=None, final_dtype=np.float64
+)
+
+
+def quantile_new_dims_func(q) -> tuple[Dim]:
+    return (Dim(name="quantile", values=q),)
+
+
+quantile = Aggregation(
+    name="quantile",
+    fill_value=dtypes.NA,
+    chunk=None,
+    combine=None,
+    final_dtype=np.float64,
+    new_dims_func=quantile_new_dims_func,
+)
+nanquantile = Aggregation(
+    name="nanquantile",
+    fill_value=dtypes.NA,
+    chunk=None,
+    combine=None,
+    final_dtype=np.float64,
+    new_dims_func=quantile_new_dims_func,
+)
+mode = Aggregation(name="mode", fill_value=dtypes.NA, chunk=None, combine=None)
+nanmode = Aggregation(name="nanmode", fill_value=dtypes.NA, chunk=None, combine=None)
 
 aggregations = {
     "any": any_,
@@ -496,6 +580,12 @@ aggregations = {
     "nanfirst": nanfirst,
     "last": last,
     "nanlast": nanlast,
+    "median": median,
+    "nanmedian": nanmedian,
+    "quantile": quantile,
+    "nanquantile": nanquantile,
+    "mode": mode,
+    "nanmode": nanmode,
 }
 
 
@@ -504,7 +594,7 @@ def _initialize_aggregation(
     dtype,
     array_dtype,
     fill_value,
-    min_count: int | None,
+    min_count: int,
     finalize_kwargs: dict[Any, Any] | None,
 ) -> Aggregation:
     if not isinstance(func, Aggregation):
@@ -529,6 +619,7 @@ def _initialize_aggregation(
 
     final_dtype = _normalize_dtype(dtype_ or agg.dtype_init["final"], array_dtype, fill_value)
     agg.dtype = {
+        "user": dtype,  # Save to automatically choose an engine
         "final": final_dtype,
         "numpy": (final_dtype,),
         "intermediate": tuple(
@@ -542,6 +633,7 @@ def _initialize_aggregation(
     }
 
     # Replace sentinel fill values according to dtype
+    agg.fill_value["user"] = fill_value
     agg.fill_value["intermediate"] = tuple(
         _get_fill_value(dt, fv)
         for dt, fv in zip(agg.dtype["intermediate"], agg.fill_value["intermediate"])
@@ -559,9 +651,6 @@ def _initialize_aggregation(
         assert isinstance(finalize_kwargs, dict)
         agg.finalize_kwargs = finalize_kwargs
 
-    if min_count is None:
-        min_count = 0
-
     # This is needed for the dask pathway.
     # Because we use intermediate fill_value since a group could be
     # absent in one block, but present in another block
@@ -569,9 +658,10 @@ def _initialize_aggregation(
     # where the identity element is 0, 1
     if min_count > 0:
         agg.min_count = min_count
-        agg.chunk += ("nanlen",)
         agg.numpy += ("nanlen",)
-        agg.combine += ("sum",)
+        if agg.chunk != (None,):
+            agg.chunk += ("nanlen",)
+            agg.combine += ("sum",)
         agg.fill_value["intermediate"] += (0,)
         agg.fill_value["numpy"] += (0,)
         agg.dtype["intermediate"] += (np.intp,)
@@ -579,7 +669,7 @@ def _initialize_aggregation(
     else:
         agg.min_count = 0
 
-    simple_combine = []
+    simple_combine: list[Callable | None] = []
     for combine in agg.combine:
         if isinstance(combine, str):
             if combine in ["nanfirst", "nanlast"]:
