@@ -2,11 +2,14 @@
 # defined in xarray
 
 import datetime
-from typing import Any, Iterable
+import importlib
+from collections.abc import Iterable
+from typing import Any, Optional
 
 import numpy as np
 import pandas as pd
 from numpy.core.multiarray import normalize_axis_index  # type: ignore[attr-defined]
+from packaging.version import Version
 
 try:
     import cftime
@@ -19,7 +22,7 @@ try:
 
     dask_array_type = dask.array.Array
 except ImportError:
-    dask_array_type = ()  # type: ignore
+    dask_array_type = ()  # type: ignore[assignment, misc]
 
 
 def asarray(data, xp=np):
@@ -81,7 +84,7 @@ class ReprObject:
 def is_scalar(value: Any, include_0d: bool = True) -> bool:
     """Whether to treat a value as a scalar.
 
-    Any non-iterable, string, or 0-D array
+    Any non-iterable, string, dict, or 0-D array
     """
     NON_NUMPY_SUPPORTED_ARRAY_TYPES = (dask_array_type, pd.Index)
 
@@ -89,7 +92,7 @@ def is_scalar(value: Any, include_0d: bool = True) -> bool:
         include_0d = getattr(value, "ndim", None) == 0
     return (
         include_0d
-        or isinstance(value, (str, bytes))
+        or isinstance(value, (str, bytes, dict))
         or not (
             isinstance(value, (Iterable,) + NON_NUMPY_SUPPORTED_ARRAY_TYPES)
             or hasattr(value, "__array_function__")
@@ -97,6 +100,20 @@ def is_scalar(value: Any, include_0d: bool = True) -> bool:
     )
 
 
+def notnull(data):
+    if not is_duck_array(data):
+        data = np.asarray(data)
+
+    scalar_type = data.dtype.type
+    if issubclass(scalar_type, (np.bool_, np.integer, np.character, np.void)):
+        # these types cannot represent missing values
+        return np.ones_like(data, dtype=bool)
+    else:
+        out = isnull(data)
+        np.logical_not(out, out=out)
+        return out
+
+
 def isnull(data):
     if not is_duck_array(data):
         data = np.asarray(data)
@@ -157,7 +174,7 @@ def datetime_to_numeric(array, offset=None, datetime_unit=None, dtype=float):
         if array.dtype.kind in "Mm":
             offset = _datetime_nanmin(array)
         else:
-            offset = min(array)
+            offset = array.min()
 
     # Compute timedelta object.
     # For np.datetime64, this can silently yield garbage due to overflow.
@@ -316,3 +333,26 @@ def nanlast(values, axis, keepdims=False):
         return np.expand_dims(result, axis=axis)
     else:
         return result
+
+
+def module_available(module: str, minversion: Optional[str] = None) -> bool:
+    """Checks whether a module is installed without importing it.
+
+    Use this for a lightweight check and lazy imports.
+
+    Parameters
+    ----------
+    module : str
+        Name of the module.
+
+    Returns
+    -------
+    available : bool
+        Whether the module is installed.
+    """
+    has = importlib.util.find_spec(module) is not None
+    if has:
+        mod = importlib.import_module(module)
+        return Version(mod.__version__) >= Version(minversion) if minversion is not None else True
+    else:
+        return False
